use mapping_layer::MappingLayer;

type UnsignedNumber = u64;
mod mapping_layer;
mod parsing;
mod range_mapping;
pub fn solve_task_1(input: &str) -> String {
    let (seeds, layers) = parsing::parse(input);
    let location = get_lowest_locations_from(&seeds, &layers);
    location.to_string()
}

fn get_lowest_locations_from(seeds: &[UnsignedNumber], layers: &[MappingLayer]) -> UnsignedNumber {
    seeds
        .into_iter()
        .map(|next_seed| location_of_one_point(*next_seed, layers))
        .min()
        .unwrap()
}

fn location_of_one_point(seed: UnsignedNumber, layers: &[MappingLayer]) -> UnsignedNumber {
    let mut current_position = seed;
    for next_layer in layers {
        let ranges = next_layer.ranges();
        let range_to_use = match ranges
            .binary_search_by_key(&current_position, |extract_source| extract_source.source())
        {
            Ok(exact) => ranges
                .get(exact)
                .expect("Exact index can not be out of bounds"),
            Err(not_exact) => {
                let not_exact = not_exact.clamp(0, ranges.len() - 1);
                let source = ranges.get(not_exact).unwrap();
                let source_pos = source.source();
                if current_position > source_pos {
                    source
                } else {
                    ranges.get(not_exact.saturating_sub(1)).unwrap()
                }
            }
        };
        current_position = range_to_use.map_position(current_position);
    }
    current_position
}

#[cfg(test)]
mod testing {
    use super::*;

    // Seed 79, soil 81, fertilizer 81, water 81, light 74, temperature 78, humidity 78, location 82.
    // Seed 14, soil 14, fertilizer 53, water 49, light 42, temperature 42, humidity 43, location 43.
    // Seed 55, soil 57, fertilizer 57, water 53, light 46, temperature 82, humidity 82, location 86.
    // Seed 13, soil 13, fertilizer 52, water 41, light 34, temperature 34, humidity 35, location 35.
    #[test]
    fn name() {
        fn assert_case(seeds: UnsignedNumber, layer: &[MappingLayer], expected: UnsignedNumber) {
            let actual = location_of_one_point(seeds, &layer);
            assert_eq!(expected, actual, "Given seeds: {}", seeds);
        }
        let input = parsing::parse(include_str!("day5/day5_example_input.txt"));
        let layers = &input.1;
        assert_case(79, layers, 82);
        assert_case(14, layers, 43);
        assert_case(55, layers, 86);
        assert_case(13, layers, 35);
    }
}
